#define _GNU_SOURCE 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <fitsio.h>
#include "SpiceUsr.h"
#define  STRLEN 100
#define  BCVLEN 5       //max number of FOV bound vectors
#define  ROOM    10

int getlos_cam_c ( SpiceInt        instid,
                  SpiceDouble     pixel[2],
                  SpiceDouble     los_vector [3] );
int get_cam_info (
                  SpiceInt        instid,
                  SpiceDouble     center[2],
                  SpiceDouble     pixels[2] );


int main(int argc, char *argv[]) {
  
  // general purpose variables
  SpiceChar   instr[STRLEN];
  SpiceDouble    pixel_dbl[2];
  SpiceDouble    image_size[2];
  SpiceDouble    image_center[2];
  SpiceInt    address;
  SpiceBoolean found, lightened;
  SpiceInt         n_return;
  
  // fits related variables
  fitsfile *incidence_fptr, *emission_fptr, *phase_fptr;
  fitsfile *lat_fptr, *lon_fptr;
  int dim = 2;
  long int naxis[2];
  long int pixel[2];
  int status = 0;
  
  // output files related
  double *incidence_img, *emission_img, *phase_img;
  double *lat_img, *lon_img, *albedo_img;
  SpiceChar   incidence_fname[STRLEN], phase_fname[STRLEN];
  SpiceChar   emission_fname[STRLEN], lat_fname[STRLEN], lon_fname[STRLEN];
  SpiceChar   base_fname_1[STRLEN];
  SpiceChar   base_fname_2[STRLEN];
  
  SpiceDouble incidence,emission;
  SpiceDouble phase,radius;
  SpiceDouble lat,lon;
  
  SpiceDouble SC_pos_bd[3],solar_pos_bd[3];
  SpiceDouble et,lt, trgepc;
  
  //getfov_c releted variables
  SpiceInt         instid = -275610; //instrument NAIF ID
  SpiceChar        shape  [STRLEN];
  SpiceChar        frame  [STRLEN]; //name of frame that FOV/boresight are defined.
  SpiceDouble      bsight [3]; //boresight vector
  SpiceInt         n_bcv; //number of FOV bound vectors
  SpiceDouble      bounds [BCVLEN][3];
  SpiceDouble      bsight_bd [3]; //boresight vector
  
  SpiceDouble rotate[3][3];
  
  SpiceDouble surface_point[3], xsrfvec[3];

  if (argc < 2) {
    printf("Usage: %s kernel [...]\n",argv[0]);
    exit(-1);
  }
  
  while (argc > 1) {
    furnsh_c (argv[1]);
    printf("%s is loaded.\n",argv[1]);
    --argc;
    ++argv;
  }
  

  if (!get_cam_info (instid, image_center, image_size)) {
    printf("Camera %d is not found.\n",instid);
    exit(0);
  }
  sprintf (instr,"FRAME_%d_NAME",instid);
  gcpool_c ( instr, 0, ROOM, STRLEN, &n_return, instr, &found );
  printf("Use %s \n",instr);
  naxis[0] = (long int)image_size[0];
  naxis[1] = (long int)image_size[1];
  
  getfov_c ( instid, BCVLEN, STRLEN, STRLEN,
          shape, frame,  bsight, &n_bcv, bounds );
      
  //Prepare large buffer
  incidence_img  = (double *)malloc(sizeof(double)*naxis[0]*naxis[1]);
  emission_img  = (double *)malloc(sizeof(double)*naxis[0]*naxis[1]);
  phase_img  = (double *)malloc(sizeof(double)*naxis[0]*naxis[1]);
  lat_img  = (double *)malloc(sizeof(double)*naxis[0]*naxis[1]);
  lon_img  = (double *)malloc(sizeof(double)*naxis[0]*naxis[1]);
  
  //Prepare base filename
  prompt_c ( "Base filename: ", STRLEN, base_fname_1 );
  
  prompt_c ( "Input time (UTC): ", STRLEN, instr );
  str2et_c (instr,&et);
  timout_c ( et, "YYYY-MM-DDTHRMNSC::UTC", STRLEN, base_fname_2 );
      
  spkpos_c ( "PLANET-C", et,   "IAU_VENUS",  "LT+S", "VENUS", SC_pos_bd,  &lt);
  spkpos_c ( "SUN", et,   "IAU_VENUS",  "LT+S", "VENUS", solar_pos_bd,  &lt);
      
  printf ("S/C position (Body-Fixed):\n");
  printf ( "    X   = %16.5f\n", SC_pos_bd[0]);
  printf ( "    Y   = %16.5f\n", SC_pos_bd[1]);
  printf ( "    Z   = %16.5f\n", SC_pos_bd[2]);
  printf ("Solar position (Body-Fixed):\n");
  printf ( "    X   = %16.5f\n", solar_pos_bd[0]);
  printf ( "    Y   = %16.5f\n", solar_pos_bd[1]);
  printf ( "    Z   = %16.5f\n", solar_pos_bd[2]);
  
  //Prepare file name
  PREP_FILE_NAME:
  sprintf(incidence_fname, "%s_%s_i.fits",base_fname_1,base_fname_2);
  sprintf(emission_fname, "%s_%s_e.fits",base_fname_1,base_fname_2);
  sprintf(phase_fname, "%s_%s_g.fits",base_fname_1,base_fname_2);
  sprintf(lat_fname, "%s_%s_lat.fits",base_fname_1,base_fname_2);
  sprintf(lon_fname, "%s_%s_lon.fits",base_fname_1,base_fname_2);
  
  //Open output files
  fits_create_file(&incidence_fptr, incidence_fname, &status);
  fits_create_file(&emission_fptr, emission_fname, &status);
  fits_create_file(&phase_fptr, phase_fname, &status);
  fits_create_file(&lat_fptr, lat_fname, &status);
  fits_create_file(&lon_fptr, lon_fname, &status);
  
  if (status != 0) {
    printf("file create error!\n");
    printf("Output filenames may not unique.\n");
    prompt_c ( "Please change Base filename: ", STRLEN, base_fname_1 );
    status = 0;
    goto PREP_FILE_NAME;
    //Back to Prepare filename then create them
  }
   
  pixel_dbl[0]=pixel_dbl[1]=511.5;
  getlos_cam_c (instid, pixel_dbl, bsight);
  mxv_c (rotate,bsight,bsight_bd);
  printf ( "The central LOS vector (body-fixed):\n");
  printf ( "    X   = %16.5f\n", bsight[0]);
  printf ( "    Y   = %16.5f\n", bsight[1]);
  printf ( "    Z   = %16.5f\n", bsight[2]);
  
  
  for (pixel[1] = 0; pixel[1] < naxis[1]; ++pixel[1]) {
    if(!(pixel[1]%100)) {
      printf ( "\rline: %d", pixel[1]);
      fflush(NULL);
    }
    for (pixel[0] = 0; pixel[0] < naxis[0]; ++pixel[0]) {
      
      pixel_dbl[0] = (SpiceDouble)pixel[0];
      pixel_dbl[1] = (SpiceDouble)pixel[1];
      getlos_cam_c (instid, pixel_dbl, bsight);
      
      //address of the pixel in the image array
      address = pixel[1]*naxis[0] + pixel[0]; 
      
      mxv_c (rotate,bsight,bsight_bd);
      
      sincpt_c ( "Ellipsoid", "VENUS", et, "IAU_VENUS", "NONE", "PLANET-C", "IAU_VENUS", bsight_bd,
                surface_point, &trgepc, xsrfvec, &found );

      if (!found) { //ɃfȂ
        incidence_img[address] = (double)(180.0);
        emission_img[address] = (double)(180.0);
        phase_img[address] = (double)(180.0);
        lat_img[address] = (double)(-91.0);
        lon_img[address] = (double)(-1.0);
      }
      
      if (found) {
        reclat_c ( surface_point, &radius, &lon, &lat);
        lat =   dpr_c()*lat;
        lon =   dpr_c()*lon;
        if (lon < 0) lon += 360.00;
        lat_img[address] = (double)lat;
        lon_img[address] = (double)lon;

        ilumin_c ( "Ellipsoid", "VENUS", et, "IAU_VENUS", "NONE", "PLANET-C", surface_point,
                  &trgepc, xsrfvec, &phase, &incidence, &emission);
        incidence_img[address] = (double)dpr_c()*incidence;
        emission_img[address] = (double)dpr_c()*emission;
        phase_img[address] = (double)dpr_c()*phase;
                
      }
    }
  }
  //Output fits files
  fits_create_img(incidence_fptr, DOUBLE_IMG, dim, naxis, &status);
  fits_write_2d_dbl(incidence_fptr, 0, naxis[0], naxis[0], naxis[1], incidence_img, &status);
  fits_close_file(incidence_fptr, &status);
  
  fits_create_img(emission_fptr, DOUBLE_IMG, dim, naxis, &status);
  fits_write_2d_dbl(emission_fptr, 0, naxis[0], naxis[0], naxis[1], emission_img, &status);
  fits_close_file(emission_fptr, &status);
  
  fits_create_img(phase_fptr, DOUBLE_IMG, dim, naxis, &status);
  fits_write_2d_dbl(phase_fptr, 0, naxis[0], naxis[0], naxis[1], phase_img, &status);
  fits_close_file(phase_fptr, &status);
  
  fits_create_img(lat_fptr, DOUBLE_IMG, dim, naxis, &status);
  fits_write_2d_dbl(lat_fptr, 0, naxis[0], naxis[0], naxis[1], lat_img, &status);
  fits_close_file(lat_fptr, &status);
  
  fits_create_img(lon_fptr, DOUBLE_IMG, dim, naxis, &status);
  fits_write_2d_dbl(lon_fptr, 0, naxis[0], naxis[0], naxis[1], lon_img, &status);
  fits_close_file(lon_fptr, &status);
}

